package edu.hm.ee.rlc.parser;

/**
 * "Verbesserte Queue" zum Parsen der Eingabe in Form 
 * meiner RLC-Schaltungsbeschreibung. Es gibt equivalente
 * Funktionen zur Queue, naemlich poll und peek. Ausserdem
 * gibt es Funkionen, mit denen gleichzeitig auf Gleichheit
 * getestet wird.
 * 
 * Leerzeichen werden grundsaetzlich ueberlesen, ausser wenn
 * sie in den Vergleich-Strings vorkommen.
 * 
 * Saemtliche Funktionen geben null zurueck, wenn das Ende des
 * zu parsenden Strings erreicht ist!
 * 
 * @author jakob190590
 *
 */
public class RLCInputStream {
	
	private StringBuffer sb; // string buffer
	private int index = 0;
	private int lastIndex = 0;
	
	public RLCInputStream() {
		sb  = new StringBuffer();
	}
	
	public RLCInputStream(String input) {
		sb  = new StringBuffer(input);
	}
	
	public RLCInputStream(RLCInputStream otherStream) {
		sb = new StringBuffer(otherStream.sb);
	}
	
	/**
	 * Zeichen nachsehen.
	 * @return
	 */
	public Character peek() {
		skipBlanks();
		return (index < sb.length()) ? sb.charAt(index) : null;
	}
	
	/**
	 * Zeichen/String nachsehen.
	 * @param count
	 * @return
	 */
	public String peek(int count) {
		skipBlanks();
		if (index < sb.length()) {
			if (index + count >= sb.length())
				return sb.substring(index, sb.length());
			else
				return sb.substring(index, index + count);
		} else return null;
	}
	
	/**
	 * Zeichen vergleichen.
	 * @param c
	 * @return
	 */
	public boolean peekEqual(char c) {
		Character ch = peek();
		return ch != null && ch == c; 
	}
	
	/**
	 * Zeichen/String vergleichen.
	 * @param s
	 * @return
	 */
	public boolean peekEqual(String s) {
		String st = peek(s.length());
		return st != null && st.equals(s);
	}
	
	/**
	 * Ein Zeichen abholen.
	 * @return
	 */
	public Character poll() {
		Character c = peek();
		if (index < sb.length()) {
			updateLastIndex();
			index++;
		}
		return c;
	}
	
	/**
	 * Zeichen/String abholen.
	 * @param count Anzahl der angeforderten Zeichen
	 * @return wenn count zu gross: Rest, sonst geforderte Anzahl von Zeichen
	 */
	public String poll(int count) {
		String s = peek(count);
		if (index < sb.length()) {
			updateLastIndex();
			index += count;
		}
		return s;
	}
	
	/**
	 * Momentante Position im String beim Parsen.
	 * Maximal Laenge des String, also eins nach letztem gueltigem Index.
	 * Minimal 0, also Anfang.
	 * @return
	 */
	public int getCurrentPos() {
		return (index > sb.length()) ? sb.length() : index;
	}
	
	public int getLastPos() {
		return lastIndex;
	}
	
	/**
	 * Zeichen nur bei Gleichheit abholen.
	 * @param c Vergleichszeichen
	 * @return true, wenn beide Zeichen gleich sind
	 */
	public boolean pollIfEqual(char c) {
		Character ch = peek();
		if (ch != null && ch == c) {
			poll();
			return true;
		} else return false;
	}
	
	/**
	 * Zeichen/String nur bei Gleichheit abholen.
	 * @param s Vergleichsstring
	 * @return true, bei Gleichheit
	 */
	public boolean pollIfEqual(String s) {
		String st = peek(s.length());
		if (st != null && st.equals(s)) {
			poll(s.length());
			return true;
		} else return false;
	}
	
	/**
	 * Ende des zu parsenden Strings erreicht?
	 * @return
	 */
	public boolean eol() {
		skipBlanks();
		return index >= sb.length();
	}
	
	/**
	 * Stream zuruecksetzen, so dass er wieder von vorne mit dem Lesen beginnt.
	 * index und lastIndex werden beide auf 0 zurueckgesetzt.
	 */
	public void reset() {
		index = lastIndex = 0;		
	}
	
	/**
	 * Cursor von der aktuellen Position um offset verschieben.
	 * @param offset
	 * @return neue Position im String, aequivalent zu getCurrentPos().
	 */
	public int seek(int offset) {
		int newIndex = getCurrentPos() + offset;
		updateLastIndex();
		index = (newIndex < 0) ? 0 : newIndex;
		
		return getCurrentPos();
	}
	
	/**
	 * Hilfsfunktion zum ueberspringen der (fuehrenden) Leerzeichen.
	 */
	private void skipBlanks() {
		while (index < sb.length() && sb.charAt(index) == ' ')
			index++;
	}
	
	/**
	 * Setzt lastIndex auf den aktuellen Index. Muss aufgerufen werden,
	 * bevor der aktuelle Index veraendert wird. lastIndex gibt bei einem
	 * Fehler die Position an, an der der Fehler auftrat.
	 */
	private void updateLastIndex() {
		lastIndex = getCurrentPos();
	}	

	@Override
	public String toString() {
		return sb.toString();
	}

}
