package edu.hm.ee.rlc.parser;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import edu.hm.ee.rlc.comps.IndukL;
import edu.hm.ee.rlc.comps.KapC;
import edu.hm.ee.rlc.comps.OhmR;
import edu.hm.ee.rlc.comps.ParSchaltg;
import edu.hm.ee.rlc.comps.SerSchaltg;
import edu.hm.ee.rlc.comps.Zweipol;

/**
 * Dieser Parser ist bis jetzt nicht besonders tolerant:
 * 
 * - Bei Wert fuer Bauelement kein Plus (wg. + fuer Reihenschaltung) - Folgende
 * SI-Prefixe: f, p, n, u/, m, k, M/meg, G, T, P (case-sensitive) - Keine
 * Widerstandsbenennung wie R1, C2, ... (wird eh ned gespeichert) - Bei
 * Widerstand R ist keine Masseinheit erlaubt (einfach weglassen)
 * 
 * Gueltige Beispiele: ( R=4e3 + C=2e-9 ) || L=40e-3 // Bauelementspezifikation
 * durch Formelzeichen ( 4e3 + 2e-9 F ) || 40e-3 H // Bauelementspezifikation
 * durch Einheit ( 4k + 2 nF ) || 40 mH // SI-Prefixe moeglich (sogar in
 * Kombination mit Exponentialschreibweise) (gemischt ist auch moeglich)
 * 
 * Hinweis: Parallelschaltung hat hoehere Prioritaet, wird also vor
 * Reihenschaltung behandelt.
 * 
 * Wichtig: Siehe RLC-Schaltungsbeschreibung.txt
 * 
 * @author jakob190590
 * 
 */
public class StandardParser implements RLCParser {

	private RLCInputStream stream;
	
	private Map<String, Double> identifier = new HashMap<String, Double>();

	@Override
	public synchronized Zweipol parse(String in) throws ParseException {
		// Eingabe als "Queue-Stream" bereitstellen
		stream = new RLCInputStream(in);

		Zweipol r = reihe();

		if (!stream.eol()) // stream NICHT komplett ausgelesen -> hinweis auf fehler
			throw new ParseException("Syntax error, parsing stopped.", stream.getCurrentPos());

		return r;
	}

	private Zweipol reihe() throws ParseException {
		List<Zweipol> list = new ArrayList<Zweipol>();

		list.add(paral());

		while (stream.pollIfEqual(REIHE_OP))
			list.add(paral());

		// Liste der Reihen-Elemente, von hinten her abraeumen
		// und immer zwei zamfassen bis nur noch eines uebrig ist
		int i = list.size() - 1;
		while (i-- >= 1)
			list.add(new SerSchaltg(list.remove(i), list.remove(i)));

		return list.get(0);
	}

	private Zweipol paral() throws ParseException {
		List<Zweipol> list = new ArrayList<Zweipol>();

		list.add(ausdruck());

		while (stream.pollIfEqual(PARAL_OP)) {
			list.add(ausdruck());
		}

		// Liste der Parallel-Elemente, von hinten her abraeumen
		// und immer zwei zamfassen bis nur noch eines uebrig ist
		int i = list.size() - 1;
		while (i-- >= 1)
			list.add(new ParSchaltg(list.remove(i), list.remove(i)));

		return list.get(0);
	}

	private Zweipol ausdruck() throws ParseException {
		if (stream.eol())
			throw new ParseException("ausdruck expected, but EOL found.",
					stream.getCurrentPos());

		Zweipol ret;

		char c = stream.peek();
		if (c == '(') {
			stream.poll();
			ret = reihe();

			// Klammer zu ) oder zumindest EOL erwartet:
			Character rightPerenthesis = stream.poll();
			if (rightPerenthesis != null && rightPerenthesis != ')')
				throw new ParseException("')' or EOL expected, but '"
						+ rightPerenthesis + "' found.", stream.getLastPos());
		} else
			ret = bauelement();

		return ret;
	}

	private Zweipol bauelement() throws ParseException {
		if (stream.eol())
			throw new ParseException("bauelement expected, but EOL found.",
					stream.getCurrentPos());
		// TODO Faktor ablesen z.B. R=5k || (2 R + 3R)
		/*
		 * Problem: Faktor nur Faktor, wenn dahinter R, L, C oder G kommt 
		 * z.B. an einem peek(count=30), und diesen String bis R, L, C, G in Zahl umwandeln
		 * (drauf achten, dass kein + dabei is!)
		 * dann poll(zahl.length()) und weiter gehts unten!
		 * und der faktor muss dann noch bauelement(char elem) uebergeben werden!
		 */
		double faktor = faktor();
		
		Zweipol ret;

		char c = stream.peek();
		switch (c) {
		// Formelzeichen am Anfang, z.B. R= oder R1
		case 'R':
		case 'L':
		case 'C':
		case 'G':
			stream.poll();
			ret = bauelement(c, faktor);
			break;

		default:

			double wert = wert() * siprefix();

			ret = makeBauelement(masseinheitToGroesse(masseinheit()), wert);
			break;
		}

		return ret;

	}

	/**
	 * Diese Funktion gibt den Faktor vor einem symbolischen Namen zurueck. Falls das bauelement
	 * gar keinen symolischen Namen beinhaltet wird 1 zurueckgegeben, weil es in diesem Fall
	 * auch keinen Faktor geben kann (Faktoren gibt es nur in Kombination mit symbolischen Namen).
	 * 
	 * Wenn es keinen Faktor gibt wird der Hauptstream nicht veraendert, sodass anschliessend ganz 
	 * normal verfahren werden kann. Wenn es aber einen gibt wird der Hauptstream so veraendert,
	 * dass anschliessend auch ganz normal weiterverfahren werden kann (es wird einfach der Faktor
	 * "weggelesen" sodass nur noch z.B. R=2k da steht).
	 * @return Der Faktor fuer das Bauelement bzw. 1 falls nicht vorhanden.
	 * @throws ParseException
	 */
	private double faktor() throws ParseException {
		double ret = 1;
		
		// temp-stream erstellen, der an der aktuellen position vom Hauptstream einsetzt 
		// und bearbeitet werden kann, ohne dass der Hauptstream veraendert wird!
		RLCInputStream probe = new RLCInputStream(stream);
		probe.seek(stream.getCurrentPos());

		String fakt = "";
		
		Character c; // Zahl einlesen, diesmal einfache Zahl, nicht in Exponentialschreibweise
		while ((c = probe.peek()) != null && (c == '.' || Character.isDigit(c)))
			fakt += probe.poll();
		
		c = probe.peek();
		if (fakt.length() > 0 && c != null && (c == 'R' || c == 'L' || c == 'C' || c == 'G'))
			try {
				ret = Double.parseDouble(fakt);
				stream.poll(probe.getCurrentPos() - stream.getCurrentPos());
			} catch (NumberFormatException e) {
				throw new ParseException("No valid faktor: " + fakt, stream.getCurrentPos());
			}
		
		return ret;
	}

	private Zweipol bauelement(char elem, double faktor) throws ParseException {
		
		String identif = elem + name();
		Double wert = identifier.get(identif);
		if (wert != null) {
			// diesen bezeichner gibt es schon
			
			if (stream.peekEqual('='))
				throw new ParseException("Identifier '" + identif + "' is already defined. " +
						"Use but not redefine it!", stream.getCurrentPos());
			
			// Wert mit faktor multiplizieren (z.B. 3 R1 bedeutet 3*R1 und R1 ist ein zuvor definierter Widerstand)
			wert *= faktor;
			
		} else {
			// wert == null, d.h. diesen bezeichner gibt es noch nicht.

			Character c = stream.poll();
			// auf Formelzeichen R, L, C, G muss '=' folgen!
			if (c == null)
				throw new ParseException("'=' expected, but EOL found.",
						stream.getLastPos() + 1);
			else if (c != '=')
				throw new ParseException("'=' expected, but '" + c + "' found.",
						stream.getLastPos());
	
			wert = wert() * siprefix();
	
			Character unit = masseinheit(); // angegebene masseinheit ermitteln
											// (null, wenn nicht angegeben)
			if (unit != null && elem != masseinheitToGroesse(unit))
				// einheit angegeben aber keine uebereinstimmung mit zuvor
				// angegebener groesse (durch z.B. 'R=')
				throw new ParseException("Invalid masseinheit for bauelement "
						+ elem + ": '" + unit + "'.", stream.getLastPos());
			
			// bezeichner (z.B. R1) mit seinem Wert zur Map hinzufuegen!
			identifier.put(identif, wert);

			// Wert mit faktor multiplizieren (z.B. 3 R = 2k bedeutet 3*R mit R = 2k)
			// muss gemachtwerden, nachdem der wert in der hashmap eingetragen wird!
			wert *= faktor;
		}
		
		return makeBauelement(elem, wert);
	}

	private String name() {
		String s = "";
		Character c;
		while ((c = stream.peek()) != null && (Character.isLetterOrDigit(c) || c == '_')) {
			s += c;
			stream.poll();
		}
		
		return s;
	}

	private double wert() throws ParseException {
		// Wert einlesen - (unnoetiges) Plus als Vorzeichen ist verboten, weil
		// das auch schon Reihenschaltung bedeutet.
		String s = "";

		// wenn '+' offensichtlich als vorzeichen fuer wert verwendet wird:
		if (stream.peekEqual('+'))
			throw new ParseException(
					"'+' as algebraic sign is not permitted due to ambiguity (reihe).",
					stream.getCurrentPos());

		Character c;
		while ((c = stream.peek()) != null) {
			if (Character.isDigit(c) || c == 'e' || c == 'E' || c == '.'
					|| c == '-') { // gueltige Zeichen fuer Gleitkommazahl
				s += stream.poll();

				// wenn '+' offensichtlich als vorzeichen fuer exponent
				// verwendet wird:
				if (c == 'e' || c == 'E') {
					if (stream.peekEqual('+'))
						throw new ParseException(
								"'+' as algebraic sign for exponent is not permitted due to ambiguity (reihe).",
								stream.getCurrentPos());
				}

			} else
				break;
		}

		double d;
		try {
			d = Double.parseDouble(s);
			if (d < 0) // nur Positive Bauelementwerte!
				throw new ParseException("wert must not be less than 0.", stream.getCurrentPos() - s.length());
			
		} catch (NumberFormatException e) {
			throw new ParseException("Invalid wert for bauelement" + (s.isEmpty() ? "" : ": '" + s
					+ "'."), stream.getCurrentPos() - s.length());
		}

		return d;
	}

	private double siprefix() {
		double k = 1;

		if (!stream.eol()) {

			// si-prefixe mit mehreren Zeichen muessen zuerst abgefragt werden,
			// wenn sie naemlich mit dem gleichen buchstaben beginnen wie andere!
			
			// 'meg' abfragen - wie bei SPICE
			// vorteil: alle BENOETIGTEN si-prefixe bestehen aus kleinbuchstaben!
			if (stream.pollIfEqual("meg"))
				k = 1e6;
			else {
			
				char c = stream.peek();
		
				boolean doPoll = true;
				switch (c) {
				case 'f':
					k = 1e-15;
					break;
				case 'p':
					k = 1e-12;
					break;
				case 'n':
					k = 1e-9;
					break;
				case 'u':
				case '':
					k = 1e-6;
					break;
				case 'm':
					k = 1e-3;
					break;
				case 'k':
					k = 1e3;
					break;
				case 'M':
					k = 1e6;
					break;
				case 'G':
					k = 1e9;
					break;
				case 'T':
					k = 1e12;
					break;
				case 'P':
					k = 1e15;
					break;
		
				default:
					doPoll = false;
					break;
				}
		
				if (doPoll)
					stream.poll();
			}
		
		}
		return k;
	}

	/**
	 * Stellt die Masseinheit fest. Gibt 'F', 'H', 'S' (Farad, Henry, Siemens)
	 * oder null (als Standard) fuer Ohm zurueck!
	 * 
	 * @return
	 */
	private Character masseinheit() {
		Character einheit = null; // default, ohne einheit: meist Widerstand R

		// Fuer Widerstand R gibt es hier keine Einheit, Ohm is bloed, weils
		// mehr als ein Zeichen is, deswegen faellts ganz weg.
		if (stream.pollIfEqual('F'))
			einheit = 'F';
		else if (stream.pollIfEqual('H'))
			einheit = 'H';
		else if (stream.pollIfEqual('S'))
			einheit = 'S';

		return einheit;
	}

	/**
	 * Wandelt eine Masseinheit in das Symbol fuer die entsprechende
	 * physikalische Groesse um. Ist die Einheit null oder ungueltig, so wird
	 * Widerstand R als Groesse angenommen!
	 * 
	 * @param einheit
	 * @return
	 */
	private Character masseinheitToGroesse(Character einheit) {
		if (einheit == null)
			return 'R';
		switch (einheit) {
		case 'S': // Siemens -> Leitwert = 1 / Widerstand
			return 'G';
		case 'H': // Henry -> Induktivitaet L
			return 'L';
		case 'F': // Farad -> Kapazitaet C
			return 'C';
		default: // wird nie eintreten, default ist trotzdem Widerstand R
			return 'R';
		}
	}
	
	/**
	 * Ein Bauelement anhand des Formelzeichens und des Wertes erstellen.
	 * Der Aufrufer muss sicherstellen, dass ein gueltiges Formelzeichen uebergeben wird.
	 * @param elem Formelzeichen/Symbol fuer Bauelement (R, L, C, G)
	 * @param val Bauelementwert
	 * @return
	 */
	private Zweipol makeBauelement(char elem, double val) {
		// TODO bauelement mit namen erstellen!
		switch (elem) {
		case 'R':
			return new OhmR(val);
		case 'G':
			return new OhmR(1 / val);
		case 'L':
			return new IndukL(val);
		case 'C':
			return new KapC(val);
		default:
			return null;
		}
	}
	
	
	
	public void clearIdentifier() {
		identifier.clear();
	}
	
	public Double putIdentifier(String identif, double value) {
		return identifier.put(identif, value);
	}
}
